/* Generated by Cython 0.29.21 */

#ifndef __PYX_HAVE__asynctnt__iproto__protocol
#define __PYX_HAVE__asynctnt__iproto__protocol

#include "Python.h"
struct C_TntFields;

/* "asynctnt/iproto/schema.pxd":4
 * 
 * 
 * cdef public class TntFields [object C_TntFields, type C_TntFields_Type]:             # <<<<<<<<<<<<<<
 *     cdef:
 *         list _names  # contains only field names ('f1', 'f2', ...)
 */
struct C_TntFields {
  PyObject_HEAD
  struct __pyx_vtabstruct_8asynctnt_6iproto_8protocol_TntFields *__pyx_vtab;
  PyObject *_names;
  PyObject *_mapping;
};

#ifndef __PYX_HAVE_API__asynctnt__iproto__protocol

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#ifndef DL_IMPORT
  #define DL_IMPORT(_T) _T
#endif

__PYX_EXTERN_C DL_IMPORT(PyTypeObject) C_TntFields_Type;

#endif /* !__PYX_HAVE_API__asynctnt__iproto__protocol */

/* WARNING: the interface of the module init function changed in CPython 3.5. */
/* It now returns a PyModuleDef instance instead of a PyModule instance. */

#if PY_MAJOR_VERSION < 3
PyMODINIT_FUNC initprotocol(void);
#else
PyMODINIT_FUNC PyInit_protocol(void);
#endif

#endif /* !__PYX_HAVE__asynctnt__iproto__protocol */
